# coding: utf-8
import collections
import re

from django.contrib.admindocs.utils import trim_docstring
from django.core.cache import cache
from django.core.exceptions import ImproperlyConfigured
from django.utils import inspect
from django.utils import six

from modernrpc.compat import standardize_strings
from modernrpc.conf import settings
from modernrpc.exceptions import RPCUnknownMethod, AuthenticationFailed, RPCInvalidParams
from modernrpc.utils import ensure_sequence, get_modernrpc_logger

# Special constant meaning "all protocols" or "all entry points"
ALL = "__all__"

# Protocols identifiers
JSONRPC_PROTOCOL = '__json_rpc'
XMLRPC_PROTOCOL = '__xml_rpc'

# Keys used in kwargs dict given to RPC methods
REQUEST_KEY = 'request'
ENTRY_POINT_KEY = 'entry_point'
PROTOCOL_KEY = 'protocol'
HANDLER_KEY = 'handler'

logger = get_modernrpc_logger(__name__)
registry = {}


class RPCMethod(object):

    def __init__(self, func):

        # Store the reference to the registered function
        self.function = func

        # @rpc_method decorator parameters
        self._external_name = getattr(func, 'modernrpc_name', func.__name__)
        self.entry_point = getattr(func, 'modernrpc_entry_point')
        self.protocol = getattr(func, 'modernrpc_protocol')
        self.str_standardization = getattr(func, 'str_standardization')
        self.str_std_encoding = getattr(func, 'str_standardization_encoding')
        # Authentication related attributes
        self.predicates = getattr(func, 'modernrpc_auth_predicates', None)
        self.predicates_params = getattr(func, 'modernrpc_auth_predicates_params', ())

        # List method's positional arguments
        # We can't use django.utils.inspect.get_func_args() with Python 2, because this function remove the first
        # argument in returned list. This is supposed to remove the first 'self' argument, but doesn't fork well
        # for global functions.
        # For Python 2, we will prefer django.utils.inspect.getargspec(func)[0]. This will work as expected, even if
        # the function has been removed in Django 2.0, since Django 2 doesn't work with Python 2
        self.args = inspect.get_func_args(func) if six.PY3 else inspect.getargspec(func)[0]
        # Does the method accept additional kwargs dict?
        self.accept_kwargs = inspect.func_accepts_kwargs(func)

        # Contains the signature of the method, as returned by "system.methodSignature"
        self.signature = []
        # Contains the method's docstring, in HTML form
        self.html_doc = ''
        # Contains doc about arguments and their type. We store this in an ordered dict, so the args documentation
        # keep the order defined in docstring
        self.args_doc = collections.OrderedDict()
        # Contains doc about return type and return value
        self.return_doc = {}

        # Docstring parsing
        self.raw_docstring = self.parse_docstring(func.__doc__)
        self.html_doc = self.raw_docstring_to_html(self.raw_docstring)

    @property
    def name(self):
        return self._external_name

    def __repr__(self):
        return 'RPC Method ' + self.name

    def __eq__(self, other):
        return \
            self.function == other.function and \
            self._external_name == other._external_name and \
            self.entry_point == other.entry_point and \
            self.protocol == other.protocol and \
            self.predicates == other.predicates and \
            self.predicates_params == other.predicates_params

    def parse_docstring(self, content):
        """
        Parse the given full docstring, and extract method description, arguments, and return documentation.

        This method try to find arguments description and types, and put the information in "args_doc" and "signature"
        members. Also parse return type and description, and put the information in "return_doc" member.
        All other lines are added to the returned string
        :param content: The full docstring
        :type content: str
        :return: The parsed method description
        :rtype: str
        """

        if not content:
            return

        raw_docstring = ''
        # We use the helper defined in django admindocs app to remove indentation chars from docstring,
        # and parse it as title, body, metadata. We don't use metadata for now.
        docstring = trim_docstring(content)

        # Define regular expression used to parse docstring
        PARAM_REXP = r'^:param ([\w]+):\s?(.*)'
        RETURN_REXP = r'^:return:\s?(.*)'
        PARAM_TYPE_REXP = r'^:type ([\w]+):\s?(.*)'
        RETURN_TYPE_REXP = r'^:rtype:\s?(.*)'

        for line in docstring.split('\n'):

            # Empty line
            if not line:
                raw_docstring += '\n'
                continue

            param_match = re.match(PARAM_REXP, line)
            if param_match:
                param_name, description = param_match.group(1, 2)
                if param_name == 'kwargs':
                    continue
                doc = self.args_doc.get(param_name, {})
                doc['text'] = description
                self.args_doc[param_name] = doc
                continue

            param_type_match = re.match(PARAM_TYPE_REXP, line)
            if param_type_match:
                param_name, param_type = param_type_match.group(1, 2)
                if param_name == 'kwargs':
                    continue
                doc = self.args_doc.get(param_name, {})
                doc['type'] = param_type
                self.args_doc[param_name] = doc
                self.signature.append(param_type)
                continue

            return_match = re.match(RETURN_REXP, line)
            if return_match:
                return_description = return_match.group(1)
                self.return_doc['text'] = return_description
                continue

            return_type_match = re.match(RETURN_TYPE_REXP, line)
            if return_type_match:
                return_description = return_type_match.group(1)
                self.return_doc['type'] = return_description
                self.signature.insert(0, return_description)
                continue

            # Line doesn't match with known args/return regular expressions,
            # add the line to raw help text
            raw_docstring += line + '\n'
        return raw_docstring

    @staticmethod
    def raw_docstring_to_html(docstring):

        if not docstring:
            return ''

        if settings.MODERNRPC_DOC_FORMAT.lower() in ('rst', 'reStructred', 'reStructuredText'):
            from docutils.core import publish_parts
            return publish_parts(docstring, writer_name='html')['body']

        elif settings.MODERNRPC_DOC_FORMAT.lower() in ('md', 'markdown'):
            import markdown
            return markdown.markdown(docstring)

        return "<p>{}</p>".format(docstring.replace('\n\n', '</p><p>').replace('\n', ' '))

    def check_permissions(self, request):
        """Call the predicate(s) associated with the RPC method, to check if the current request
        can actually call the method.
        Return a boolean indicating if the method should be executed (True) or not (False)"""

        if not self.predicates:
            return True

        # All registered authentication predicates must return True
        return all(
            predicate(request, *self.predicates_params[i])
            for i, predicate in enumerate(self.predicates)
        )

    def execute(self, *args, **kwargs):
        """Call the function encapsulated by the current instance"""

        if six.PY2:
            args = standardize_strings(args, strtype=self.str_standardization, encoding=self.str_std_encoding)

        # Call the rpc method, as standard python function
        return self.function(*args, **kwargs)

    def available_for_protocol(self, protocol):
        """Check if the current function can be executed from a request defining the given protocol"""
        if self.protocol == ALL or protocol == ALL:
            return True

        return protocol in ensure_sequence(self.protocol)

    def available_for_entry_point(self, entry_point):
        """Check if the current function can be executed from a request to the given entry point"""
        if self.entry_point == ALL or entry_point == ALL:
            return True

        return entry_point in ensure_sequence(self.entry_point)

    def is_valid_for(self, entry_point, protocol):
        """Check if the current function can be executed from a request to the given entry point
        and with the given protocol"""
        return self.available_for_entry_point(entry_point) and self.available_for_protocol(protocol)

    def is_available_in_json_rpc(self):
        """Shortcut checking if the current method can be executed on JSONRPC protocol.
        Used in HTML documentation to easily display protocols supported by a RPC method"""
        return self.available_for_protocol(JSONRPC_PROTOCOL)

    def is_available_in_xml_rpc(self):
        """Shortcut checking if the current method can be executed on XMLRPC protocol.
        Used in HTML documentation to easily display protocols supported by a RPC method"""
        return self.available_for_protocol(XMLRPC_PROTOCOL)

    # Helpers to simplify templates generation
    def is_doc_available(self):
        """Returns True if a textual documentation is available for this method"""
        return bool(self.html_doc)

    def is_return_doc_available(self):
        """Returns True if this method's return is documented"""
        return bool(self.return_doc and (self.return_doc.get('text') or self.return_doc.get('type')))

    def is_args_doc_available(self):
        """Returns True if any of the method's arguments is documented"""
        return self.args_doc

    def is_any_doc_available(self):
        """Returns True if there is a textual documentation or a documentation on arguments or return of the method."""
        return self.is_args_doc_available() or self.is_return_doc_available() or self.is_doc_available()


def reset_registry():
    """Flush the registry from any previously registered RPC method"""
    registry.clear()


def register_rpc_method(func):
    """
    Register a function to be available as RPC method.

    The given function will be inspected to find external_name, protocol and entry_point values set by the decorator
    @rpc_method.
    :param func: A function previously decorated using @rpc_method
    :return: The name of registered method
    """
    if not getattr(func, 'modernrpc_enabled', False):
        raise ImproperlyConfigured('Error: trying to register {} as RPC method, but it has not been decorated.'
                                   .format(func.__name__))

    # Define the external name of the function
    name = getattr(func, 'modernrpc_name', func.__name__)

    logger.info('Register RPC method {}'.format(name))

    if name.startswith('rpc.'):
        raise ImproperlyConfigured('According to RPC standard, method names starting with "rpc." are reserved for '
                                   'system extensions and must not be used. See '
                                   'http://www.jsonrpc.org/specification#extensions for more information.')

    # Encapsulate the function in a RPCMethod object
    method = RPCMethod(func)

    # Ensure method names are unique in the registry
    existing_method = get_method(method.name, ALL, ALL)
    if existing_method is not None:
        # Trying to register many times the same function is OK, because if a method is decorated
        # with @rpc_method(), it could be imported in different places of the code
        if method == existing_method:
            return method.name
        # But if we try to use the same name to register 2 different methods, we
        # must inform the developer there is an error in the code
        else:
            raise ImproperlyConfigured("A RPC method with name {} has already been registered".format(method.name))

    # Store the method
    registry[method.name] = method

    return method.name


def get_all_method_names(entry_point=ALL, protocol=ALL, sort_methods=False):
    """Return the names of all RPC methods registered supported by the given entry_point / protocol pair"""

    method_names = [
        name for name, method in registry.items() if method.is_valid_for(entry_point, protocol)
    ]

    if sort_methods:
        method_names = sorted(method_names)

    return method_names


def get_all_methods(entry_point=ALL, protocol=ALL, sort_methods=False):
    """Return a list of all methods in the registry supported by the given entry_point / protocol pair"""

    if sort_methods:
        return [method for (_, method) in sorted(registry.items()) if method.is_valid_for(entry_point, protocol)]

    return registry.values()


def get_method(name, entry_point, protocol):
    """Retrieve a method from the given name"""

    if name in registry and registry[name].is_valid_for(entry_point, protocol):
        return registry[name]

    return None


def rpc_method(func=None, name=None, entry_point=ALL, protocol=ALL,
               str_standardization=settings.MODERNRPC_PY2_STR_TYPE,
               str_standardization_encoding=settings.MODERNRPC_PY2_STR_ENCODING):
    """
    Mark a standard python function as RPC method.

    All arguments are optional

    :param func: A standard function
    :param name: Used as RPC method name instead of original function name
    :param entry_point: Default: ALL. Used to limit usage of the RPC method for a specific set of entry points
    :param protocol: Default: ALL. Used to limit usage of the RPC method for a specific protocol (JSONRPC or XMLRPC)
    :param str_standardization: Default: settings.MODERNRPC_PY2_STR_TYPE. Configure string standardization on python 2.
    Ignored on python 3.
    :param str_standardization_encoding: Default: settings.MODERNRPC_PY2_STR_ENCODING. Configure the encoding used
    to perform string standardization conversion
    :type name: str
    :type entry_point: str
    :type protocol: str
    :type str_standardization: type str or unicode
    :type str_standardization_encoding: str
    """

    def decorated(_func):

        _func.modernrpc_enabled = True
        _func.modernrpc_name = name or _func.__name__
        _func.modernrpc_entry_point = entry_point
        _func.modernrpc_protocol = protocol
        _func.str_standardization = str_standardization
        _func.str_standardization_encoding = str_standardization_encoding

        return _func

    # If @rpc_method() is used with parenthesis (with or without argument)
    if func is None:
        return decorated

    # If @rpc_method is used without parenthesis
    return decorated(func)


def clean_old_cache_content():
    """Clean CACHE data from old versions of django-modern-rpc"""
    cache.delete('__rpc_registry__', version=1)


class RPCRequest(object):
    """
    Encapsulate a RPC request.

    Instances of this class are used to call a RPC methods the same way from anywhere.
    """

    def __init__(self, method_name, args=None, kwargs=None):
        self.method_name = method_name
        self.args = args or []
        self.kwargs = kwargs or {}

    def execute(self, handler):
        """
        Process RPC request, call the corresponding RPC Method and return the result.

        Raise RPCUnknownMethod, AuthenticationFailed, RPCInvalidParams or any Exception sub-class.
        """

        rpc_method = get_method(self.method_name, handler.entry_point, handler.protocol)

        if not rpc_method:
            logger.warning('Unknown RPC method: {}'.format(self.method_name))
            raise RPCUnknownMethod(self.method_name)

        logger.debug('Check authentication / permissions for method {} and user {}'
                     .format(self.method_name, handler.request.user))

        if not rpc_method.check_permissions(handler.request):
            logger.info('Call to {} forbidden by authentication system.'.format(self.method_name))
            raise AuthenticationFailed()

        logger.debug('RPC method {} will be executed'.format(self.method_name))

        # Build args & kwargs for procedure execution
        args, kwargs = self.args, self.kwargs

        # If the RPC method needs to access some internals, update kwargs dict
        if rpc_method.accept_kwargs:
            kwargs.update({
                REQUEST_KEY: handler.request,
                ENTRY_POINT_KEY: handler.entry_point,
                PROTOCOL_KEY: handler.protocol,
                HANDLER_KEY: handler,
            })

        logger.debug('Params: args = {} - kwargs = {}'.format(args, kwargs))

        try:
            # Call the python function associated with the RPC method name
            return rpc_method.execute(*args, **kwargs)

        except TypeError as e:
            # If given arguments cannot be transmitted properly to python function,
            # raise an Invalid Params exceptions
            raise RPCInvalidParams(str(e))



# coding: utf-8
import inspect
import warnings
from importlib import import_module

import django.core.checks
import django.utils.version
from django.apps import AppConfig

from modernrpc.conf import settings
from modernrpc.core import register_rpc_method, reset_registry, clean_old_cache_content


@django.core.checks.register()
def check_required_settings_defined(app_configs, **kwargs):  # noqa
    result = []
    if not settings.MODERNRPC_METHODS_MODULES:
        result.append(
            django.core.checks.Warning(
                'settings.MODERNRPC_METHODS_MODULES is not set, django-modern-rpc cannot locate your RPC methods.',
                hint='Please define MODERNRPC_METHODS_MODULES in your settings.py to indicate modules containing your '
                     'methods. See http://django-modern-rpc.rtfd.io/en/latest/basic_usage/methods_registration.html '
                     'for more info',
                obj=settings,
                id='modernrpc.E001',
            )
        )
    return result


class ModernRpcConfig(AppConfig):

    name = 'modernrpc'
    verbose_name = 'Django Modern RPC'

    def ready(self):
        self.rpc_methods_registration()

    @staticmethod
    def rpc_methods_registration():
        """Look into each module listed in settings.MODERNRPC_METHODS_MODULES, import each module and register
        functions annotated with @rpc_method decorator in the registry"""

        # In previous version, django-modern-rpc used the django cache system to store methods registry.
        # It is useless now, so clean the cache from old data
        clean_old_cache_content()
        # For security (and unit tests), make sure the registry is empty before registering rpc methods
        reset_registry()

        if not settings.MODERNRPC_METHODS_MODULES:
            # settings.MODERNRPC_METHODS_MODULES is undefined or empty, but we already notified user
            # with check_required_settings_defined() function. See http://docs.djangoproject.com/en/1.10/topics/checks/
            return

        # Lookup content of MODERNRPC_METHODS_MODULES, and add the module containing system methods
        for module_name in settings.MODERNRPC_METHODS_MODULES + ['modernrpc.system_methods']:

            try:
                # Import the module in current scope
                rpc_module = import_module(module_name)

            except ImportError:
                msg = 'Unable to load module "{}" declared in settings.MODERNRPC_METHODS_MODULES. Please ensure ' \
                      'it is available and doesn\'t contains any error'.format(module_name)
                warnings.warn(msg, category=Warning)
                continue

            # Lookup all global functions in module
            for _, func in inspect.getmembers(rpc_module, inspect.isfunction):
                # And register only functions with attribute 'modernrpc_enabled' defined to True
                if getattr(func, 'modernrpc_enabled', False):
                    register_rpc_method(func)



# coding: utf-8

from django.conf import settings as user_settings

from modernrpc.conf import default_settings


class ModernRpcSettings(object):

    def __getattr__(self, item):

        try:
            # First, try to retrieve setting from project level settings module
            return getattr(user_settings, item)
        except AttributeError:
            # Fallback: return the default value, provided by django-modern-rpc
            return getattr(default_settings, item)


settings = ModernRpcSettings()



# coding: utf-8
import json

from django.http.response import HttpResponse
from django.utils.module_loading import import_string

from modernrpc.conf import settings
from modernrpc.core import JSONRPC_PROTOCOL
from modernrpc.exceptions import RPCInternalError, RPCInvalidRequest, RPCParseError, RPCException
from modernrpc.handlers.base import RPCHandler
from modernrpc.core import RPCRequest

try:
    # Python 3
    from json.decoder import JSONDecodeError
except ImportError:
    # Python 2: json.loads will raise a ValueError when loading json
    JSONDecodeError = ValueError


class JSONRPCBatchResult(object):
    def __init__(self):
        self.results = []


class JSONRPCHandler(RPCHandler):

    protocol = JSONRPC_PROTOCOL

    def __init__(self, request, entry_point):
        super(JSONRPCHandler, self).__init__(request, entry_point)
        self.request_id = None

    @staticmethod
    def valid_content_types():
        return [
            'application/json',
        ]

    def loads(self, str_data):
        try:
            decoder = import_string(settings.MODERNRPC_JSON_DECODER)
            return json.loads(str_data, cls=decoder)

        except JSONDecodeError as e:
            raise RPCParseError(str(e))

    def dumps(self, obj):
        try:
            encoder = import_string(settings.MODERNRPC_JSON_ENCODER)
            return json.dumps(obj, cls=encoder)
        except Exception:
            raise RPCInternalError('Unable to serialize result as valid JSON')

    def process_request(self):

        encoding = self.request.encoding or 'utf-8'
        payload = self.loads(self.request.body.decode(encoding))

        if isinstance(payload, dict):

            # Store current request id, or None if request is a notification
            self.request_id = payload.get('id')
            return self.process_single_request(payload)

        elif isinstance(payload, (list, tuple)):

            batch_result = JSONRPCBatchResult()

            for single_payload in payload:
                try:
                    try:
                        request_id = single_payload.get('id')

                    except AttributeError:
                        request_id = None
                        raise RPCInvalidRequest()

                    result = self.process_single_request(single_payload)

                    if request_id:
                        # As stated in documentation:
                        # "A Response object SHOULD exist for each Request object, except that there SHOULD NOT be any
                        # Response objects for notifications."
                        batch_result.results.append(self.json_success_response(result, override_id=request_id))

                except RPCException as e:
                    batch_result.results.append(self.json_error_response(e, override_id=request_id))

                except Exception as e:
                    rpc_exception = RPCInternalError(str(e))
                    batch_result.results.append(self.json_error_response(rpc_exception, override_id=request_id))

            return batch_result

        else:
            raise RPCInvalidRequest()

    def process_single_request(self, payload):

        if 'jsonrpc' not in payload:
            raise RPCInvalidRequest('Missing parameter "jsonrpc"')

        elif 'method' not in payload:
            raise RPCInvalidRequest('Missing parameter "method"')

        if payload['jsonrpc'] != '2.0':
            raise RPCInvalidRequest('The attribute "jsonrpc" must contains "2.0"')

        params = payload.get('params')

        if isinstance(params, (list, tuple)):
            rpc_request = RPCRequest(payload['method'], args=params)

        elif isinstance(params, dict):
            rpc_request = RPCRequest(payload['method'], kwargs=params)

        else:
            rpc_request = RPCRequest(payload['method'])

        return rpc_request.execute(self)

    @staticmethod
    def json_http_response(data, http_response_cls=HttpResponse):
        response = http_response_cls(data)
        response['Content-Type'] = 'application/json'
        return response

    def json_success_response(self, data, override_id=None):

        if not (override_id or self.request_id):
            return None

        return {
            'id': override_id or self.request_id,
            'jsonrpc': '2.0',
            'result': data,
        }

    def result_success(self, data):

        result = None

        if isinstance(data, JSONRPCBatchResult):

            # Result data for Batch requests is ready to dump, no need to insert it in a
            # response payload
            # As stated in standard:
            # "If there are no Response objects contained within the Response array as it is to be sent to the client,
            # the server MUST NOT return an empty Array and should return nothing at all."
            result = data.results or None

        elif self.request_id is not None:
            result = self.json_success_response(data)

        if result is None:
            # Nothing should be returned when JSON-RPC request is a notification or a batch
            # of notification only requests
            return HttpResponse(status=204)

        return self.json_http_response(self.dumps(result))

    def json_error_response(self, exception, override_id=None):

        result = {
            'id': override_id or self.request_id,
            'jsonrpc': '2.0',
            'error': {
                'code': exception.code,
                'message': exception.message,
            }
        }

        if exception.data:
            result['error']['data'] = exception.data

        return result

    def result_error(self, exception, http_response_cls=HttpResponse):
        result = self.json_error_response(exception)
        return self.json_http_response(self.dumps(result), http_response_cls=http_response_cls)



# coding: utf-8
from modernrpc.exceptions import RPCInvalidRequest


class RPCHandler(object):

    protocol = None

    def __init__(self, request, entry_point):
        self.request = request
        self.entry_point = entry_point

    def loads(self, str_data):
        """Convert serialized string data to valid Python data, depending on current handler protocol"""
        raise NotImplementedError("You must override loads()")

    def dumps(self, obj):
        """Convert Python data to serialized form, according to current handler protocol."""
        raise NotImplementedError("You must override dumps()")

    @staticmethod
    def valid_content_types():
        raise NotImplementedError("You must override valid_content_types()")

    def can_handle(self):
        # Get the content-type header from incoming request. Method differs depending on current Django version
        try:
            # Django >= 1.10
            content_type = self.request.content_type
        except AttributeError:
            # Django up to 1.9
            content_type = self.request.META['CONTENT_TYPE']

        if not content_type:
            # We don't accept a request with missing Content-Type request
            raise RPCInvalidRequest('Missing header: Content-Type')

        return content_type.lower() in self.valid_content_types()

    def process_request(self):
        """
        Parse self.request to extract payload. Parse it to retrieve RPC call information, and
        execute the corresponding RPC Method. At any time, raise an exception when detecting error.
        :return: The result of RPC Method execution
        """
        raise NotImplementedError("You must override process_request()")

    def result_success(self, data):
        """Return a HttpResponse instance containing the result payload for the given data"""
        raise NotImplementedError("You must override result_success()")

    def result_error(self, exception, http_response_cls=None):
        """Return a HttpResponse instance containing the result payload for the given exception"""
        raise NotImplementedError("You must override result_error()")



# coding: utf-8
from django.utils import six

from modernrpc.conf import settings


def generic_convert_string(v, from_type, to_type, encoding):
    """Generic method to convert from a unicode to str or from str to unicode. Works only with Python 2"""

    if from_type == str and isinstance(v, str):
        return unicode(v, encoding)

    elif from_type == unicode and isinstance(v, unicode):
        return v.encode(encoding)

    elif isinstance(v, (list, tuple, set)):
        return type(v)([generic_convert_string(element, from_type, to_type, encoding) for element in v])

    elif isinstance(v, dict):
        return {k: generic_convert_string(v, from_type, to_type, encoding) for k, v in v.iteritems()}

    return v


def standardize_strings(arg, strtype=settings.MODERNRPC_PY2_STR_TYPE, encoding=settings.MODERNRPC_PY2_STR_ENCODING):
    assert six.PY2, "This function should be used with Python 2 only"

    if not strtype:
        return arg

    if strtype == str or strtype == 'str':
        # We want to convert from unicode to str
        return generic_convert_string(arg, unicode, str, encoding)

    elif strtype == unicode or strtype == 'unicode':
        # We want to convert from str to unicode
        return generic_convert_string(arg, str, unicode, encoding)

    raise TypeError('standardize_strings() called with an invalid strtype: "{}"'.format(repr(strtype)))



# coding: utf-8
import base64

from django.contrib.auth import authenticate, login
from django.contrib.auth.models import AnonymousUser
from django.utils import six

from modernrpc import auth
from modernrpc.auth import user_is_authenticated


def http_basic_auth_check_user(request, user_validation_func, *args):
    """
    Inspect the given request and extract the user / password from HTTP_AUTHORIZATION header.
    Then try to use it to authenticate against Django User model. If user can be authenticated, it is returned.
    If no user is found, an anonymous user object is returned.
    :param request: The request to inspect
    :param user_validation_func:
    :return:
    """
    return user_validation_func(http_basic_auth_get_user(request), *args)


def http_basic_auth_get_user(request):
    """Inspect the given request to find a logged user. If not found, the header HTTP_AUTHORIZATION
    is read for 'Basic Auth' login and password, and try to authenticate against default UserModel.
    Always return a User instance (possibly anonymous, meaning authentication failed)"""

    try:
        # If standard auth middleware already authenticated a user, use it
        if user_is_authenticated(request.user):
            return request.user
    except AttributeError:
        pass

    # This was grabbed from https://www.djangosnippets.org/snippets/243/
    # Thanks to http://stackoverflow.com/a/1087736/1887976
    if 'HTTP_AUTHORIZATION' in request.META:
        auth_data = request.META['HTTP_AUTHORIZATION'].split()
        if len(auth_data) == 2 and auth_data[0].lower() == "basic":
            uname, passwd = base64.b64decode(auth_data[1]).decode('utf-8').split(':')
            login(request, authenticate(username=uname, password=passwd))

    # In all cases, return the current request's user (may be anonymous user if no login succeed)
    try:
        return request.user
    except AttributeError:
        return AnonymousUser()


# Decorator
def http_basic_auth_login_required(func=None):
    """Decorator. Use it to specify a RPC method is available only to logged users"""

    wrapper = auth.set_authentication_predicate(http_basic_auth_check_user, [auth.user_is_authenticated])

    # If @http_basic_auth_login_required() is used (with parenthesis)
    if func is None:
        return wrapper

    # If @http_basic_auth_login_required is used without parenthesis
    return wrapper(func)


# Decorator
def http_basic_auth_superuser_required(func=None):
    """Decorator. Use it to specify a RPC method is available only to logged superusers"""

    wrapper = auth.set_authentication_predicate(http_basic_auth_check_user, [auth.user_is_superuser])

    # If @http_basic_auth_superuser_required() is used (with parenthesis)
    if func is None:
        return wrapper

    # If @http_basic_auth_superuser_required is used without parenthesis
    return wrapper(func)


# Decorator
def http_basic_auth_permissions_required(permissions):
    """Decorator. Use it to specify a RPC method is available only to logged users with given permissions"""

    if isinstance(permissions, six.string_types):
        return auth.set_authentication_predicate(http_basic_auth_check_user, [auth.user_has_perm, permissions])
    else:
        return auth.set_authentication_predicate(http_basic_auth_check_user, [auth.user_has_all_perms, permissions])


# Decorator
def http_basic_auth_any_of_permissions_required(permissions):
    """Decorator. Use it to specify a RPC method is available only to logged users with given permissions"""
    return auth.set_authentication_predicate(http_basic_auth_check_user, [auth.user_has_any_perm, permissions])


# Decorator
def http_basic_auth_group_member_required(groups):
    """Decorator. Use it to specify a RPC method is available only to logged users with given permissions"""

    if isinstance(groups, six.string_types):
        # Check user is in a group
        return auth.set_authentication_predicate(http_basic_auth_check_user, [auth.user_in_group, groups])
    else:
        # Check user is in many group
        return auth.set_authentication_predicate(http_basic_auth_check_user, [auth.user_in_any_group, groups])


# Decorator
def http_basic_auth_all_groups_member_required(groups):
    """Decorator. Use it to specify a RPC method is available only to logged users with given permissions"""

    return auth.set_authentication_predicate(http_basic_auth_check_user, [auth.user_in_all_groups, groups])



# coding: utf-8

# List of python modules containing RPC methods.
MODERNRPC_METHODS_MODULES = []

# Decoder class used to convert python data to JSON
MODERNRPC_JSON_DECODER = 'json.decoder.JSONDecoder'
# Encoder class used to convert JSON data to python values
MODERNRPC_JSON_ENCODER = 'django.core.serializers.json.DjangoJSONEncoder'


# Control how builtin types are handled by XML-RPC serializer and deserializer
MODERNRPC_XMLRPC_USE_BUILTIN_TYPES = True
# Control how XML-RPC serializer will handle None values. If set to True, None values will be converted to <nil>
MODERNRPC_XMLRPC_ALLOW_NONE = True
# Configure the default encoding used by XML encoder
MODERNRPC_XMLRPC_DEFAULT_ENCODING = None

# List of handler classes used by default in any ``RPCEntryPoint`` instance
MODERNRPC_HANDLERS = [
    'modernrpc.handlers.JSONRPCHandler',
    'modernrpc.handlers.XMLRPCHandler',
]

# Default name associated with anonymous ``RPCEntryPoint``
MODERNRPC_DEFAULT_ENTRYPOINT_NAME = '__default_entry_point__'

# Configure the format of the docstring used to document your RPC methods.
# Possible values are: '', 'rst' or 'md'
MODERNRPC_DOC_FORMAT = ''

MODERNRPC_PY2_STR_TYPE = None
MODERNRPC_PY2_STR_ENCODING = 'UTF-8'



# coding: utf-8
import datetime

from django.utils.six.moves import xmlrpc_client


def get_builtin_date(date, date_format="%Y-%m-%dT%H:%M:%S", raise_exception=False):
    """
    Try to convert a date to a builtin instance of ``datetime.datetime``.
    The input date can be a ``str``, a ``datetime.datetime``, a ``xmlrpc.client.Datetime`` or a ``xmlrpclib.Datetime``
    instance. The returned object is a ``datetime.datetime``.

    :param date: The date object to convert.
    :param date_format: If the given date is a str, format is passed to strptime to parse it
    :param raise_exception: If set to True, an exception will be raised if the input string cannot be parsed
    :return: A valid ``datetime.datetime`` instance
    """
    if isinstance(date, datetime.datetime):
        # Default XML-RPC handler is configured to decode dateTime.iso8601 type
        # to builtin datetime.datetim instance
        return date
    elif isinstance(date, xmlrpc_client.DateTime):
        # If constant settings.MODERNRPC_XMLRPC_USE_BUILTIN_TYPES has been set to True
        # the date is decoded as DateTime object
        return datetime.datetime.strptime(date.value, "%Y%m%dT%H:%M:%S")
    else:
        # If date is given as str (or unicode for python 2)
        # This is the normal behavior for JSON-RPC
        try:
            return datetime.datetime.strptime(date, date_format)
        except ValueError:
            if raise_exception:
                raise
            else:
                return None



# coding: utf-8
from modernrpc.core import ENTRY_POINT_KEY, PROTOCOL_KEY, get_method, rpc_method, get_all_method_names, RPCRequest, \
    HANDLER_KEY, XMLRPC_PROTOCOL
from modernrpc.exceptions import RPCInvalidParams, RPCException, RPC_INTERNAL_ERROR


@rpc_method(name='system.listMethods')
def __system_listMethods(**kwargs):
    """Returns a list of all methods available in the current entry point"""
    entry_point = kwargs.get(ENTRY_POINT_KEY)
    protocol = kwargs.get(PROTOCOL_KEY)

    return get_all_method_names(entry_point, protocol, sort_methods=True)


@rpc_method(name='system.methodSignature')
def __system_methodSignature(method_name, **kwargs):
    """
    Returns an array describing the signature of the given method name.

    The result is an array with:
     - Return type as first elements
     - Types of method arguments from element 1 to N
    :param method_name: Name of a method available for current entry point (and protocol)
    :param kwargs:
    :return: An array describing types of return values and method arguments
    """
    entry_point = kwargs.get(ENTRY_POINT_KEY)
    protocol = kwargs.get(PROTOCOL_KEY)

    method = get_method(method_name, entry_point, protocol)
    if method is None:
        raise RPCInvalidParams('The method {} is not found in the system. Unable to retrieve signature.')
    return method.signature


@rpc_method(name='system.methodHelp')
def __system_methodHelp(method_name, **kwargs):
    """
    Returns the documentation of the given method name.

    :param method_name: Name of a method available for current entry point (and protocol)
    :param kwargs:
    :return: Documentation text for the RPC method
    """
    entry_point = kwargs.get(ENTRY_POINT_KEY)
    protocol = kwargs.get(PROTOCOL_KEY)

    method = get_method(method_name, entry_point, protocol)
    if method is None:
        raise RPCInvalidParams('The method {} is not found in the system. Unable to retrieve method help.')
    return method.html_doc


@rpc_method(name='system.multicall', protocol=XMLRPC_PROTOCOL)
def __system_multiCall(calls, **kwargs):
    """
    Call multiple RPC methods at once.

    :param calls: An array of struct like {"methodName": string, "params": array }
    :param kwargs: Internal data
    :type calls: list
    :type kwargs: dict
    :return:
    """
    if not isinstance(calls, list):
        raise RPCInvalidParams('system.multicall first argument should be a list, {} given.'.format(type(calls)))

    handler = kwargs.get(HANDLER_KEY)

    results = []

    for call in calls:
        rpc_request = RPCRequest(call['methodName'], args=call.get('params'))

        try:
            result = rpc_request.execute(handler)

            # From https://mirrors.talideon.com/articles/multicall.html:
            # "Notice that regular return values are always nested inside a one-element array. This allows you to
            # return structs from functions without confusing them with faults."
            results.append([result])

        except RPCException as e:
            results.append({
                'faultCode': e.code,
                'faultString': e.message,
            })
        except Exception as e:
            results.append({
                'faultCode': RPC_INTERNAL_ERROR,
                'faultString': str(e),
            })

    return results



# coding: utf-8

from django.core.exceptions import ImproperlyConfigured
from django.http.response import HttpResponse, HttpResponseForbidden
from django.utils.decorators import method_decorator
from django.utils.module_loading import import_string
from django.views.decorators.csrf import csrf_exempt
from django.views.generic.base import View, TemplateView

from modernrpc.conf import settings
from modernrpc.core import ALL, get_all_methods
from modernrpc.exceptions import RPCInternalError, RPCException, AuthenticationFailed
from modernrpc.utils import ensure_sequence, get_modernrpc_logger

logger = get_modernrpc_logger(__name__)


class RPCEntryPoint(TemplateView):
    """
    This is the main entry point class. It inherits standard Django View class.
    """

    template_name = 'modernrpc/doc_index.html'

    entry_point = settings.MODERNRPC_DEFAULT_ENTRYPOINT_NAME
    protocol = ALL
    enable_doc = False
    enable_rpc = True

    def __init__(self, **kwargs):
        super(RPCEntryPoint, self).__init__(**kwargs)

        if not self.get_handler_classes():
            raise ImproperlyConfigured("At least 1 handler must be instantiated.")

        # Copy static list http_method_names locally (in instance), so we can dynamically customize it
        self.http_method_names = list(View.http_method_names)

        # Customize allowed HTTP methods name to forbid access to GET when this EntryPoint
        # must not display docs...
        if not self.enable_doc:
            self.http_method_names.remove('get')

        # ... and also forbid access to POST when this EntryPoint must not support RPC request (docs only view)
        if not self.enable_rpc:
            self.http_method_names.remove('post')

        logger.debug('RPC entry point "{}" initialized'.format(self.entry_point))

    # This disable CSRF validation for POST requests
    @method_decorator(csrf_exempt)
    def dispatch(self, request, *args, **kwargs):
        """Overrides the default dispatch method, to disable CSRF validation on POST requests. This
        is mandatory to ensure RPC calls wil be correctly handled"""
        return super(RPCEntryPoint, self).dispatch(request, *args, **kwargs)

    def get_handler_classes(self):
        """Return the list of handlers to use when receiving RPC requests."""

        handler_classes = [import_string(handler_cls) for handler_cls in settings.MODERNRPC_HANDLERS]

        if self.protocol == ALL:
            return handler_classes
        else:
            return [cls for cls in handler_classes if cls.protocol in ensure_sequence(self.protocol)]

    def post(self, request, *args, **kwargs):
        """
        Handle a XML-RPC or JSON-RPC request.

        :param request: Incoming request
        :param args: Additional arguments
        :param kwargs: Additional named arguments
        :return: A HttpResponse containing XML-RPC or JSON-RPC response, depending on the incoming request
        """

        logger.debug('RPC request received on entry point "{}"'.format(self.entry_point))

        for handler_cls in self.get_handler_classes():

            handler = handler_cls(request, self.entry_point)

            try:
                if not handler.can_handle():
                    continue

                logger.debug('Request will be handled by {}'.format(handler_cls.__name__))

                result = handler.process_request()

                return handler.result_success(result)

            except AuthenticationFailed as e:
                # Customize HttpResponse instance used when AuthenticationFailed was raised
                return handler.result_error(e, HttpResponseForbidden)

            except RPCException as e:
                logger.warning('RPC exception raised: {}'.format(str(e)))
                return handler.result_error(e)

            except Exception as e:
                logger.error('Exception raised from a RPC method: {}'.format(str(e)))
                return handler.result_error(RPCInternalError(str(e)))

        logger.error('Received a request impossible to handle with available handlers.')

        return HttpResponse('Unable to handle your request. Please ensure you called the right entry point. If not, '
                            'this could be a server error.')

    def get_context_data(self, **kwargs):
        """Update context data with list of RPC methods of the current entry point.
        Will be used to display methods documentation page"""
        ctx = super(RPCEntryPoint, self).get_context_data(**kwargs)
        ctx.update({
            'methods': get_all_methods(self.entry_point, sort_methods=True),
        })
        return ctx



import logging

from django.utils import six


def logger_has_handlers(logger):
    """Since Python 2 doesn't provide Logger.hasHandlers(), we have to
    perform the lookup by ourself."""
    if six.PY3:
        return logger.hasHandlers()
    else:
        c = logger
        rv = False
        while c:
            if c.handlers:
                rv = True
                break
            if not c.propagate:
                break
            else:
                c = c.parent
        return rv


def ensure_sequence(element):
    """Ensure the given argument is a sequence object (tuple, list). If not, return a list with it."""
    if isinstance(element, (tuple, list)):
        return element

    return [element]


def get_modernrpc_logger(name):
    """Get a logger from default logging manager. If no handler is associated, add a default NullHandler"""
    logger = logging.getLogger(name)
    if not logger_has_handlers(logger):
        # If logging is not configured in the current project, configure this logger to discard all logs messages.
        # This will prevent the 'No handlers could be found for logger XXX' error on Python 2,
        # and avoid redirecting errors to the default 'lastResort' StreamHandler on Python 3
        logger.addHandler(logging.NullHandler())
    return logger



# coding: utf-8
import xml

from django.http.response import HttpResponse
from django.utils.six.moves import xmlrpc_client

from modernrpc.conf import settings
from modernrpc.core import XMLRPC_PROTOCOL
from modernrpc.exceptions import RPCParseError, RPCInvalidRequest, RPCInternalError
from modernrpc.handlers.base import RPCHandler
from modernrpc.core import RPCRequest


class XMLRPCHandler(RPCHandler):

    protocol = XMLRPC_PROTOCOL

    def __init__(self, request, entry_point):
        super(XMLRPCHandler, self).__init__(request, entry_point)
        # Marshaller is used to dumps data into valid XML-RPC response. See self.dumps() for more info
        self.marshaller = xmlrpc_client.Marshaller(encoding=settings.MODERNRPC_XMLRPC_DEFAULT_ENCODING,
                                                   allow_none=settings.MODERNRPC_XMLRPC_ALLOW_NONE)

    @staticmethod
    def valid_content_types():
        return [
            'text/xml',
        ]

    def loads(self, str_data):
        try:
            try:
                # Python 3
                return xmlrpc_client.loads(str_data, use_builtin_types=settings.MODERNRPC_XMLRPC_USE_BUILTIN_TYPES)
            except TypeError:
                # Python 2
                return xmlrpc_client.loads(str_data, use_datetime=settings.MODERNRPC_XMLRPC_USE_BUILTIN_TYPES)

        except xml.parsers.expat.ExpatError as e:
            raise RPCParseError(e)

        except Exception as e:
            raise RPCInvalidRequest(e)

    def dumps(self, obj):

        try:
            # Marshaller has a specific handling of Fault instance. It is given without modification
            if isinstance(obj, xmlrpc_client.Fault):
                return self.marshaller.dumps(obj)

            # xmlrpc_client.Marshaller expects a list of objects to dumps.
            # It will output a '<params></params>' block and loops onto given objects to inject, for each one,
            # a '<param><value><type>X</type></value></param>' block.
            # This is not the return defined in XML-RPC standard, see http://xmlrpc.scripting.com/spec.html:
            # "The body of the response is a single XML structure, a <methodResponse>, which can contain
            # a single <params> which contains a single <param> which contains a single <value>."
            #
            # So, to make sure the return value always contain a single '<param><value><type>X</type></value></param>',
            # we dumps it as an array of a single value.
            return self.marshaller.dumps([obj])

        except Exception:
            raise RPCInternalError('Unable to serialize result as valid XML')

    def process_request(self):

        encoding = self.request.encoding or 'utf-8'
        data = self.request.body.decode(encoding)

        params, method_name = self.loads(data)

        if method_name is None:
            raise RPCInvalidRequest('Missing methodName')

        rpc_request = RPCRequest(method_name, args=params)

        return rpc_request.execute(self)

    @staticmethod
    def xml_http_response(data, http_response_cls=HttpResponse):
        response = http_response_cls(data)
        response['Content-Type'] = 'text/xml'
        return response

    def result_success(self, data):

        raw_response = '<?xml version="1.0"?>'
        raw_response += '<methodResponse>'
        raw_response += self.dumps(data)
        raw_response += '</methodResponse>'

        return self.xml_http_response(raw_response)

    def result_error(self, exception, http_response_cls=HttpResponse):

        raw_response = '<?xml version="1.0"?>'
        raw_response += '<methodResponse>'
        raw_response += self.dumps(xmlrpc_client.Fault(exception.code, exception.message))
        raw_response += '</methodResponse>'

        return self.xml_http_response(raw_response, http_response_cls=http_response_cls)



# coding: utf-8
from distutils.version import StrictVersion

import django
from django.contrib.auth.models import Group
from django.utils import six


# Decorator
def set_authentication_predicate(predicate, params=()):
    """
    Assign a new authentication predicate to an RPC method.
    This is the most generic decorator used to implement authentication.
    Predicate is a standard function with the following signature:

    .. code:: python

       def my_predicate(request, *params):
           # Inspect request and extract required information

           if <condition>:
               # The condition to execute the method are met
               return True
           return False

    :param predicate:
    :param params:
    :return:
    """

    def wrapper(rpc_method):

        if hasattr(rpc_method, 'modernrpc_auth_predicates'):
            rpc_method.modernrpc_auth_predicates.append(predicate)
            rpc_method.modernrpc_auth_predicates_params.append(params)

        else:
            rpc_method.modernrpc_auth_predicates = [predicate]
            rpc_method.modernrpc_auth_predicates_params = [params]

        return rpc_method

    return wrapper


def user_is_authenticated(user):
    if StrictVersion(django.get_version()) < StrictVersion('1.10'):
        return user.is_authenticated()

    return user.is_authenticated


def user_is_anonymous(user):
    if StrictVersion(django.get_version()) < StrictVersion('1.10'):
        return user.is_anonymous()

    return user.is_anonymous


def user_is_superuser(user):
    return user.is_superuser


def user_has_perm(user, perm):
    return user_is_superuser(user) or user.has_perm(perm)


def user_has_all_perms(user, perms):
    """Returns True if the given user have all given permissions"""
    return user_is_superuser(user) or user.has_perms(perms)


def user_has_any_perm(user, perms):
    return user_is_superuser(user) or any(user_has_perm(user, perm) for perm in perms)


def user_in_group(user, group):
    """Returns True if the given user is in given group"""
    if isinstance(group, Group):
        return user_is_superuser(user) or group in user.groups.all()
    elif isinstance(group, six.string_types):
        return user_is_superuser(user) or user.groups.filter(name=group).exists()
    raise TypeError("'group' argument must be a string or a Group instance")


def user_in_any_group(user, groups):
    """Returns True if the given user is in at least 1 of the given groups"""
    return user_is_superuser(user) or any(user_in_group(user, group) for group in groups)


def user_in_all_groups(user, groups):
    """Returns True if the given user is in all given groups"""
    return user_is_superuser(user) or all(user_in_group(user, group) for group in groups)
